// Author: Pete Kirkham
//
// quad store utils.
//
// Some threading primitives.
//
#ifndef _THREADING_H_
#define _THREADING_H_

#include <pthread.h>

class runnable
{
public:
    virtual ~runnable () { }
    virtual void run () = 0;
};

template <class C>
class delegate : public runnable
{
public:
    typedef void (C::*f_type)();

    delegate (C* self, f_type f) : _self(self), _f(f) { }
    delegate (C& self, f_type f) : _self(&self), _f(f) { }
    
    virtual void run () 
    {
        (_self->*_f)();
    }

private:
    C*      _self;
    f_type  _f;
};

class thread
{
protected:
    thread (pthread_t id) : _id(id) { }
    
public:
    typedef void * (*start_routine)(void *);
    
    static void run (runnable& runnable)
    {
        runnable.run();
    }
    
    static void run_and_delete (runnable* t)
    {
        t->run();
        delete t;
    }

    template <class T> static thread spawn (T* t, void (T::*f)())
    {
        pthread_t   thread_id(0);
        pthread_create(&thread_id, NULL, reinterpret_cast<start_routine>(&thread::run_and_delete), reinterpret_cast<void*>(new delegate<T>(t, f)));
        return thread(thread_id);
    }

    template <class T> static thread spawn (T& t, void (T::*f)())
    {
        pthread_t   thread_id(0);
        pthread_create(&thread_id, NULL, reinterpret_cast<start_routine>(&thread::run_and_delete), reinterpret_cast<void*>(new delegate<T>(&t, f)));
        return thread(thread_id);
    }
    
    void join ()
    {
        pthread_join(_id, NULL);
    }
    
    
    
protected:
    pthread_t   _id;
};


class condition
{
public:
    condition () ;
    ~condition () ;
    
    void signal () ;
    void wait_for () ;    
    
private:
    bool            _is_signalled;
    pthread_cond_t  _cond;
    pthread_mutex_t _mutex;
};

#endif // _THREADING_H_
